package rtr;

import addr.addrIP;
import java.util.ArrayList;
import java.util.List;
import tab.tabLabel;
import tab.tabLabelNtry;
import tab.tabRoute;
import tab.tabRouteEntry;

/**
 * bgp4 update group
 *
 * @author matecsaba
 */
public class rtrBgpGroup extends rtrBgpParam {

    /**
     * group number
     */
    public final int groupNum;

    /**
     * minimum version
     */
    public int minversion;

    /**
     * willing unicast prefixes
     */
    public tabRoute<addrIP> wilUni = new tabRoute<addrIP>("tx");

    /**
     * willing multicast prefixes
     */
    public tabRoute<addrIP> wilMlt = new tabRoute<addrIP>("tx");

    /**
     * willing flowspec prefixes
     */
    public tabRoute<addrIP> wilFlw = new tabRoute<addrIP>("tx");

    /**
     * willing vpnuni prefixes
     */
    public tabRoute<addrIP> wilVpnU = new tabRoute<addrIP>("tx");

    /**
     * willing vpnmulti prefixes
     */
    public tabRoute<addrIP> wilVpnM = new tabRoute<addrIP>("tx");

    /**
     * willing vpnflow prefixes
     */
    public tabRoute<addrIP> wilVpnF = new tabRoute<addrIP>("tx");

    /**
     * willing vpls prefixes
     */
    public tabRoute<addrIP> wilVpls = new tabRoute<addrIP>("tx");

    /**
     * willing evpn prefixes
     */
    public tabRoute<addrIP> wilEvpn = new tabRoute<addrIP>("tx");

    /**
     * willing mdt prefixes
     */
    public tabRoute<addrIP> wilMdt = new tabRoute<addrIP>("tx");

    /**
     * changed unicast prefixes
     */
    public tabRoute<addrIP> chgUni = new tabRoute<addrIP>("chg");

    /**
     * changed multicast prefixes
     */
    public tabRoute<addrIP> chgMlt = new tabRoute<addrIP>("chg");

    /**
     * changed flowspec prefixes
     */
    public tabRoute<addrIP> chgFlw = new tabRoute<addrIP>("chg");

    /**
     * changed vpnuni prefixes
     */
    public tabRoute<addrIP> chgVpnU = new tabRoute<addrIP>("chg");

    /**
     * changed vpnmulti prefixes
     */
    public tabRoute<addrIP> chgVpnM = new tabRoute<addrIP>("chg");

    /**
     * changed vpnflow prefixes
     */
    public tabRoute<addrIP> chgVpnF = new tabRoute<addrIP>("chg");

    /**
     * changed vpls prefixes
     */
    public tabRoute<addrIP> chgVpls = new tabRoute<addrIP>("chg");

    /**
     * changed evpn prefixes
     */
    public tabRoute<addrIP> chgEvpn = new tabRoute<addrIP>("chg");

    /**
     * changed mdt prefixes
     */
    public tabRoute<addrIP> chgMdt = new tabRoute<addrIP>("chg");

    /**
     * local address
     */
    public addrIP localAddr;

    /**
     * type of peer
     */
    protected int peerType;

    /**
     * create group
     *
     * @param parent bgp process
     * @param num group number
     */
    public rtrBgpGroup(rtrBgp parent, int num) {
        super(parent, false);
        groupNum = num;
    }

    public void flapBgpConn() {
    }

    public void doTempCfg(String cmd, boolean negated) {
    }

    public void getConfig(List<String> l, String beg, boolean filter) {
        l.addAll(getParamCfg(beg, "group " + groupNum + " ", filter));
    }

    /**
     * get status of group
     *
     * @param l list to append
     */
    public void getStatus(List<String> l) {
        String a = "";
        for (int i = 0; i < lower.neighs.size(); i++) {
            rtrBgpNeigh ntry = lower.neighs.get(i);
            if (ntry.groupMember != groupNum) {
                continue;
            }
            a += " " + ntry.peerAddr;
        }
        for (int i = 0; i < lower.lstnNei.size(); i++) {
            rtrBgpNeigh ntry = lower.lstnNei.get(i);
            if (ntry.groupMember != groupNum) {
                continue;
            }
            a += " " + ntry.peerAddr;
        }
        l.add("peers =" + a);
        l.add("type = " + rtrBgpUtil.peerType2string(peerType));
        l.add("safi = " + mask2string(addrFams));
        l.add("local = " + localAddr);
        l.add("unicast advertise = " + wilUni.size() + ", list=" + chgUni.size());
        l.add("multicast advertise = " + wilMlt.size() + ", list=" + chgMlt.size());
        l.add("flowspec advertise = " + wilFlw.size() + ", list=" + chgFlw.size());
        l.add("vpnuni advertise = " + wilVpnU.size() + ", list=" + chgVpnU.size());
        l.add("vpnmlt advertise = " + wilVpnM.size() + ", list=" + chgVpnM.size());
        l.add("vpnflw advertise = " + wilVpnF.size() + ", list=" + chgVpnF.size());
        l.add("vpls advertise = " + wilVpls.size() + ", list=" + chgVpls.size());
        l.add("evpn advertise = " + wilEvpn.size() + ", list=" + chgEvpn.size());
        l.add("mdt advertise = " + wilMdt.size() + ", list=" + chgMdt.size());
        l.add("version = " + minversion + " of " + lower.compRound);
    }

    /**
     * get willing
     *
     * @param safi safi to query
     * @return table
     */
    public tabRoute<addrIP> getWilling(int safi) {
        if (safi == lower.afiUni) {
            return wilUni;
        }
        if (safi == lower.afiLab) {
            return wilUni;
        }
        if (safi == lower.afiMlt) {
            return wilMlt;
        }
        if (safi == lower.afiFlw) {
            return wilFlw;
        }
        if (safi == lower.afiVpnU) {
            return wilVpnU;
        }
        if (safi == lower.afiVpnM) {
            return wilVpnM;
        }
        if (safi == lower.afiVpnF) {
            return wilVpnF;
        }
        if (safi == lower.afiVpls) {
            return wilVpls;
        }
        if (safi == lower.afiEvpn) {
            return wilEvpn;
        }
        if (safi == lower.afiMdt) {
            return wilMdt;
        }
        return null;
    }

    /**
     * get changed
     *
     * @param safi safi to query
     * @return table
     */
    public tabRoute<addrIP> getChanged(int safi) {
        if (safi == lower.afiUni) {
            return chgUni;
        }
        if (safi == lower.afiLab) {
            return chgUni;
        }
        if (safi == lower.afiMlt) {
            return chgMlt;
        }
        if (safi == lower.afiFlw) {
            return chgFlw;
        }
        if (safi == lower.afiVpnU) {
            return chgVpnU;
        }
        if (safi == lower.afiVpnM) {
            return chgVpnM;
        }
        if (safi == lower.afiVpnF) {
            return chgVpnF;
        }
        if (safi == lower.afiVpls) {
            return chgVpls;
        }
        if (safi == lower.afiEvpn) {
            return chgEvpn;
        }
        if (safi == lower.afiMdt) {
            return chgMdt;
        }
        return null;
    }

    private void nextHopSelf(tabRouteEntry<addrIP> ntry) {
        ntry.nextHop = localAddr.copyBytes();
        ntry.labelRem = new ArrayList<Integer>();
        tabLabelNtry loc = ntry.labelLoc;
        if (loc == null) {
            if (ntry.rouTab == null) {
                loc = lower.fwdCore.commonLabel;
            } else {
                loc = ntry.rouTab.commonLabel;
            }
        }
        ntry.labelRem.add(loc.getValue());
    }

    /**
     * originate prefix
     *
     * @param ntry prefix
     */
    public void originatePrefix(tabRouteEntry<addrIP> ntry) {
        nextHopSelf(ntry);
        switch (peerType) {
            case rtrBgpUtil.peerExtrn:
            case rtrBgpUtil.peerServr:
                ntry.pathSeq = tabLabel.prependLabel(ntry.pathSeq, localAs);
                break;
            case rtrBgpUtil.peerCnfed:
                ntry.confSeq = tabLabel.prependLabel(ntry.confSeq, localAs);
                if (ntry.locPref == 0) {
                    ntry.locPref = 100;
                }
                break;
            case rtrBgpUtil.peerIntrn:
            case rtrBgpUtil.peerRflct:
                if (ntry.locPref == 0) {
                    ntry.locPref = 100;
                }
                break;
        }
        if ((sendCommunity & 1) == 0) {
            ntry.stdComm = null;
        }
        if ((sendCommunity & 2) == 0) {
            ntry.extComm = null;
        }
    }

    /**
     * readvertise prefix
     *
     * @param nhs set next hop self
     * @param ntry prefix
     * @return false on success, true on failure
     */
    public boolean readvertPrefix(boolean nhs, tabRouteEntry<addrIP> ntry) {
        if (!allowAsOut) {
            if (rtrBgpUtil.findIntList(ntry.pathSeq, remoteAs) >= 0) {
                return true;
            }
            if (rtrBgpUtil.findIntList(ntry.pathSet, remoteAs) >= 0) {
                return true;
            }
        }
        if (rtrBgpUtil.findIntList(ntry.stdComm, rtrBgpUtil.commNoAdvertise) >= 0) {
            return true;
        }
        switch (peerType) {
            case rtrBgpUtil.peerExtrn:
                if (rtrBgpUtil.findIntList(ntry.stdComm, rtrBgpUtil.commNoExport) >= 0) {
                    return true;
                }
                ntry.pathSeq = tabLabel.prependLabel(ntry.pathSeq, localAs);
                if (ntry.pathSeq.size() > 1) {
                    ntry.metric = 0;
                }
                ntry.locPref = 0;
                ntry.originator = null;
                ntry.clustList = null;
                ntry.confSeq = null;
                ntry.confSet = null;
                if (!nxtHopUnchgd) {
                    nextHopSelf(ntry);
                }
                break;
            case rtrBgpUtil.peerCnfed:
                if (rtrBgpUtil.findIntList(ntry.stdComm, rtrBgpUtil.commNoConfed) >= 0) {
                    return true;
                }
                switch (ntry.rouSrc) {
                    case rtrBgpUtil.peerExtrn:
                    case rtrBgpUtil.peerServr:
                        if (!nxtHopUnchgd) {
                            nextHopSelf(ntry);
                        }
                        break;
                }
                ntry.confSeq = tabLabel.prependLabel(ntry.confSeq, localAs);
                ntry.originator = null;
                ntry.clustList = null;
                break;
            case rtrBgpUtil.peerIntrn:
                switch (ntry.rouSrc) {
                    case rtrBgpUtil.peerIntrn:
                        return true;
                    case rtrBgpUtil.peerExtrn:
                    case rtrBgpUtil.peerServr:
                        if (!nxtHopUnchgd) {
                            nextHopSelf(ntry);
                        }
                        break;
                }
                break;
            case rtrBgpUtil.peerRflct:
                switch (ntry.rouSrc) {
                    case rtrBgpUtil.peerExtrn:
                    case rtrBgpUtil.peerServr:
                        if (!nxtHopUnchgd) {
                            nextHopSelf(ntry);
                        }
                        break;
                }
                break;
            case rtrBgpUtil.peerServr:
                if (rtrBgpUtil.findIntList(ntry.stdComm, rtrBgpUtil.commNoExport) >= 0) {
                    return true;
                }
                ntry.locPref = 0;
                ntry.originator = null;
                ntry.clustList = null;
                ntry.confSeq = null;
                ntry.confSet = null;
                if (!nxtHopUnchgd) {
                    nextHopSelf(ntry);
                }
                break;
        }
        if ((sendCommunity & 1) == 0) {
            ntry.stdComm = null;
            ntry.lrgComm = null;
        }
        if ((sendCommunity & 2) == 0) {
            ntry.extComm = null;
        }
        if (removePrivAsOut) {
            rtrBgpUtil.removePrivateAs(ntry.pathSeq);
            rtrBgpUtil.removePrivateAs(ntry.pathSet);
        }
        if (overridePeerOut) {
            rtrBgpUtil.replaceIntList(ntry.pathSeq, remoteAs, localAs);
            rtrBgpUtil.replaceIntList(ntry.pathSet, remoteAs, localAs);
        }
        if (nxtHopSelf) {
            nextHopSelf(ntry);
        }
        if (!accIgp) {
            ntry.accIgp = 0;
        }
        if (!traffEng) {
            ntry.bandwidth = 0;
        }
        if (nhs && ((addrFams & rtrBgpParam.mskLab) != 0) && (ntry.labelRem == null)) {
            nextHopSelf(ntry);
        }
        return false;
    }

    private void readvertTable(tabRoute<addrIP> tab, tabRoute<addrIP> cmp) {
        for (int i = 0; i < cmp.size(); i++) {
            tabRouteEntry<addrIP> ntry = cmp.get(i);
            ntry = ntry.copyBytes();
            if (readvertPrefix(true, ntry)) {
                continue;
            }
            tabRoute.addUpdatedEntry(2, tab, ntry, roumapOut, roupolOut, prflstOut);
        }
    }

    private void importTable(tabRoute<addrIP> tab, tabRoute<addrIP> imp) {
        for (int i = 0; i < imp.size(); i++) {
            tabRouteEntry<addrIP> ntry = imp.get(i);
            ntry = ntry.copyBytes();
            if (ntry.rouSrc == rtrBgpUtil.peerOriginate) {
                originatePrefix(ntry);
            } else if (readvertPrefix(false, ntry)) {
                continue;
            }
            tabRoute.addUpdatedEntry(2, tab, ntry, voumapOut, voupolOut, null);
        }
    }

    /**
     * create needed prefix list
     *
     * @param cUni unicast
     * @param cMlt multicast
     * @param cFlw flowspec
     * @param cVpnU vpn uni
     * @param cVpnM vpn multi
     * @param cVpnF vpn flow
     * @param cVpls vpls
     * @param cEvpn evpn
     * @param cMdt mdt
     */
    public void createNeeded(tabRoute<addrIP> cUni, tabRoute<addrIP> cMlt, tabRoute<addrIP> cFlw, tabRoute<addrIP> cVpnU, tabRoute<addrIP> cVpnM, tabRoute<addrIP> cVpnF, tabRoute<addrIP> cVpls, tabRoute<addrIP> cEvpn, tabRoute<addrIP> cMdt) {
        tabRoute<addrIP> nUni = new tabRoute<addrIP>("bgp");
        tabRoute<addrIP> nMlt = new tabRoute<addrIP>("bgp");
        tabRoute<addrIP> nFlw = new tabRoute<addrIP>("bgp");
        tabRoute<addrIP> nVpnU = new tabRoute<addrIP>("bgp");
        tabRoute<addrIP> nVpnM = new tabRoute<addrIP>("bgp");
        tabRoute<addrIP> nVpnF = new tabRoute<addrIP>("bgp");
        tabRoute<addrIP> nVpls = new tabRoute<addrIP>("bgp");
        tabRoute<addrIP> nEvpn = new tabRoute<addrIP>("bgp");
        tabRoute<addrIP> nMdt = new tabRoute<addrIP>("bgp");
        if (sendDefRou) {
            tabRouteEntry<addrIP> ntry = new tabRouteEntry<addrIP>();
            ntry.prefix = rtrBgpUtil.defaultRoute(lower.afiUni);
            ntry.aggrRtr = new addrIP();
            ntry.aggrRtr.fromIPv4addr(lower.routerID);
            ntry.aggrAs = localAs;
            ntry.rouSrc = rtrBgpUtil.peerOriginate;
            originatePrefix(ntry);
            tabRoute.addUpdatedEntry(2, nUni, ntry, roumapOut, roupolOut, prflstOut);
        }
        for (int i = 0; i < lower.routerRedisted.size(); i++) {
            tabRouteEntry<addrIP> ntry = lower.routerRedisted.get(i);
            ntry = ntry.copyBytes();
            ntry.rouSrc = rtrBgpUtil.peerOriginate;
            originatePrefix(ntry);
            tabRoute.addUpdatedEntry(2, nUni, ntry, roumapOut, roupolOut, prflstOut);
        }
        nMlt.mergeFrom(3, nUni, null, true);
        readvertTable(nUni, cUni);
        readvertTable(nMlt, cMlt);
        importTable(nFlw, cFlw);
        importTable(nVpnU, cVpnU);
        importTable(nVpnM, cVpnM);
        importTable(nVpnF, cVpnF);
        importTable(nVpls, cVpls);
        importTable(nEvpn, cEvpn);
        importTable(nMdt, cMdt);
        lower.routerDoAggregates(nUni, localAddr, lower.fwdCore.commonLabel, rtrBgpUtil.peerOriginate, lower.routerID, lower.localAs);
        lower.routerDoAggregates(nMlt, localAddr, lower.fwdCore.commonLabel, rtrBgpUtil.peerOriginate, lower.routerID, lower.localAs);
        wilUni = nUni;
        wilMlt = nMlt;
        wilFlw = nFlw;
        wilVpnU = nVpnU;
        wilVpnM = nVpnM;
        wilVpnF = nVpnF;
        wilVpls = nVpls;
        wilEvpn = nEvpn;
        wilMdt = nMdt;
    }

}
